import json

from py_range_parse import Range


class JsonlObjectValidator:

    def __init__(self):
        self._seen_ids = {}

    def validate(self, data: str):
        for line in data.splitlines():
            input = json.loads(line)
            self._validate_object(input)

    def _validate_object(self, input):
        input_page = input.get("page", None)
        input_id = input.get("id", None)
        self.__remember_page_id_combo(input_page, input_id)

        if input_id is not None:
            valid_range = Range(0, 254)
            if not isinstance(input_id, int) or input_id not in valid_range:
                raise AssertionError(f"Object has invalid id '{input_id}', must be in range: {valid_range}")

        input_align = input.get("align", None)
        if input_align is not None:
            if isinstance(input_align, str):
                if input_align not in ["left", "center", "right"]:
                    raise AssertionError(f"Invalid 'align' value: {input_align}")
            if isinstance(input_align, int):
                if input_align not in [0, 1, 2]:
                    raise AssertionError(f"Invalid 'align' value: {input_align}")

    def __remember_page_id_combo(self, input_page: int, input_id: int):
        key = f"p{input_page}b{input_id}"
        if key in self._seen_ids.keys():
            raise AssertionError(f"Duplicate id detected: {key}")
        else:
            self._seen_ids[key] = True



from pathlib import Path

import click

PARAM_CFG_DIR = "cfg_dir"
PARAM_OUTPUT_DIR = "output_dir"
PARAM_DEVICE = "device"
PARAM_PURGE = "purge"
PARAM_CMD = "cmd"
PARAM_PAYLOAD = "payload"

CMD_OPTION_NAMES = {
    PARAM_CFG_DIR: ["--config-dir", "-c"],
    PARAM_OUTPUT_DIR: ["--output-dir", "-o"],
    PARAM_DEVICE: ["--device", "-d"],
    PARAM_CMD: ["--command", "-C"],
    PARAM_PAYLOAD: ["--payload", "-p"],
    PARAM_PURGE: ["--purge", "-P"]
}

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option()
def cli():
    pass


def get_option_names(parameter: str) -> list:
    """
    Returns a list of all valid console parameter names for a given parameter
    :param parameter: the parameter to check
    :return: a list of all valid names to use this parameter
    """
    return CMD_OPTION_NAMES[parameter]


@cli.command(name="generate")
@click.option(*get_option_names(PARAM_CFG_DIR),
              required=True,
              type=click.Path(exists=True, path_type=Path),
              help='Root directory of your config files.')
@click.option(*get_option_names(PARAM_OUTPUT_DIR),
              required=True,
              default=Path("./output"),
              type=click.Path(path_type=Path),
              help='Root directory of where to put the generated output files.')
@click.option(*get_option_names(PARAM_DEVICE), required=False, default=None,
              help='Only generate the output for the specified device.')
def c_generate(config_dir: Path, output_dir: Path, device: str):
    """
    Generates the output files for all devices in the given config directory.
    """
    _generate(config_dir, output_dir, device)


def _generate(config_dir: Path, output_dir: Path, device: str):
    from openhasp_config_manager.manager import ConfigManager
    processor = ConfigManager(config_dir, output_dir)

    devices = processor.analyze()
    if device is not None:
        devices = list(filter(lambda x: x.name == device, devices))

    processor.process(devices)


@cli.command(name="upload")
@click.option(*get_option_names(PARAM_CFG_DIR),
              required=True,
              type=click.Path(exists=True, path_type=Path),
              help='Root directory of your config files.')
@click.option(*get_option_names(PARAM_OUTPUT_DIR),
              required=True,
              default=Path("./output"),
              type=click.Path(path_type=Path),
              help='Root directory of where the generated output files from the "generate" command are located.')
@click.option(*get_option_names(PARAM_DEVICE), required=False, default=None,
              help='Only upload the generated files for the specified device.')
@click.option(*get_option_names(PARAM_PURGE), is_flag=True,
              help='Whether to remove files from the device which are not part of the generated output.')
def c_upload(config_dir: Path, output_dir: Path, device: str, purge: bool):
    """
    Uploads the previously generated configuration to their corresponding devices.
    """
    _upload(config_dir, output_dir, device, purge)


def _upload(config_dir: Path, output_dir: Path, device: str, purge: bool):
    from openhasp_config_manager.manager import ConfigManager
    from openhasp_config_manager.uploader import ConfigUploader

    processor = ConfigManager(config_dir, output_dir)

    devices = processor.analyze()
    if device is not None:
        devices = list(filter(lambda x: x.name == device, devices))

    uploader = ConfigUploader(output_dir)
    for device in devices:
        try:
            print(f"Uploading files to device '{device.name}'...")
            uploader.upload(device, purge)
        except Exception as ex:
            print(f"Error uploading files to '{device.name}': {ex}")


@cli.command(name="deploy")
@click.option(*get_option_names(PARAM_CFG_DIR),
              required=True,
              type=click.Path(exists=True, path_type=Path),
              help='Root directory of your config files.')
@click.option(*get_option_names(PARAM_OUTPUT_DIR),
              required=True,
              default=Path("./output"),
              type=click.Path(path_type=Path),
              help='Root directory of where to put the generated output files.')
@click.option(*get_option_names(PARAM_DEVICE), required=False, default=None,
              help='Only deploy the specified device.')
@click.option(*get_option_names(PARAM_PURGE), is_flag=True,
              help='Whether to remove files from the device which are not part of the generated output.')
def c_deploy(config_dir: Path, output_dir: Path, device: str, purge: bool):
    """
    Combines the generation and upload of a configuration.
    """
    _deploy(config_dir, output_dir, device, purge)


def _reboot(config_dir, device: str):
    from openhasp_config_manager.manager import ConfigManager
    processor = ConfigManager(config_dir, Path("./nonexistent"))

    devices = processor.analyze()
    if device is not None:
        devices = list(filter(lambda x: x.name == device, devices))

    from openhasp_config_manager.openhasp import OpenHaspClient
    client = OpenHaspClient()

    for device in devices:
        client.reboot(device)


def _reload(config_dir: Path, device: str):
    from openhasp_config_manager.manager import ConfigManager
    processor = ConfigManager(config_dir, Path("./nonexistent"))

    devices = processor.analyze()
    if device is not None:
        devices = list(filter(lambda x: x.name == device, devices))

    from openhasp_config_manager.openhasp import OpenHaspClient
    client = OpenHaspClient()

    for device in devices:
        client.command(device, "clearpage", "all")
        client.command(device, "run", "L:/boot.cmd")


def _deploy(config_dir: Path, output_dir: Path, device: str, purge: bool):
    _generate(config_dir, output_dir, device)
    _upload(config_dir, output_dir, device, purge)
    # _cmd(config_dir, device="touch_down_1", command="reboot", payload="")
    # _reload(config_dir, device)
    _reboot(config_dir, device)


@cli.command(name="cmd")
@click.option(*get_option_names(PARAM_CFG_DIR),
              required=True,
              type=click.Path(exists=True, path_type=Path),
              help='Root directory of your config files.')
@click.option(*get_option_names(PARAM_DEVICE),
              required=True,
              help='Device to send the command to.')
@click.option(*get_option_names(PARAM_CMD),
              required=True,
              help='The name of the command.')
@click.option(*get_option_names(PARAM_PAYLOAD),
              required=False,
              default="",
              help='Command payload.')
def c_cmd(config_dir: Path, device: str, command: str, payload: str):
    """
    Sends a command request to a device.
    """
    _cmd(config_dir, device, command, payload)


def _cmd(config_dir: Path, device: str, command: str, payload: str):
    from openhasp_config_manager.manager import ConfigManager
    processor = ConfigManager(config_dir, Path("./nonexistent"))

    devices = processor.analyze()
    if device is not None:
        devices = list(filter(lambda x: x.name == device, devices))

    from openhasp_config_manager.openhasp import OpenHaspClient
    client = OpenHaspClient()

    for device in devices:
        client.command(device, command, payload)


if __name__ == '__main__':
    cli()



import json
import shutil
from pathlib import Path
from typing import List

from openhasp_config_manager.const import COMMON_FOLDER_NAME, DEVICES_FOLDER_NAME
from openhasp_config_manager.model import Component, Config, Device, OpenhaspConfigManagerConfig, MqttConfig, \
    HttpConfig, GuiConfig, HaspConfig, DeviceConfig, ScreenConfig
from openhasp_config_manager.processing import DeviceProcessor
from openhasp_config_manager.processing.jsonl import JsonlObjectProcessor
from openhasp_config_manager.processing.variables import VariableManager
from openhasp_config_manager.validation import DeviceValidator, JsonlObjectValidator

CONFIG_FILE_NAME = "config.json"


class ConfigManager:

    def __init__(self, cfg_root: Path, output_root: Path):
        self._cfg_root = cfg_root
        self._output_root = output_root

        self._variable_manager = VariableManager(cfg_root)

    def analyze(self) -> List[Device]:
        """
        Analyze the files in cfg_root
        :return: list of devices
        """
        print(f"Analyzing config files in '{self._cfg_root}'...")
        return self._analyze(self._cfg_root, self._output_root)

    def _analyze(self, cfg_dir_root: Path, output_dir_root: Path) -> List[Device]:
        result: List[Device] = []

        common_components_path = Path(cfg_dir_root, COMMON_FOLDER_NAME)
        common_components = self._read_components(common_components_path, prefix="common")

        devices_path = Path(cfg_dir_root, DEVICES_FOLDER_NAME)
        if not devices_path.exists():
            raise RuntimeError(
                f"No '{DEVICES_FOLDER_NAME}' sub-folder found in '{cfg_dir_root}'. Please create it and move your "
                f"device configuration files there.")
        for device_path in devices_path.iterdir():
            if not device_path.is_dir():
                continue

            device_output_dir = Path(output_dir_root, device_path.name)
            try:
                config = self._read_config(device_path)
            except Exception as ex:
                print(f"Error reading config '{device_path}': {ex}")
                raise ex

            device_components = self._analyze_device(device_path)

            device = Device(
                path=device_path,
                name=device_path.name,
                components=common_components + device_components,
                config=config,
                output_dir=device_output_dir,
            )

            result.append(device)

        return result

    def _analyze_device(self, device_cfg_dir_root: Path) -> List[Component]:
        return self._read_components(device_cfg_dir_root)

    @staticmethod
    def _read_components(path: Path, prefix: str = "") -> List[Component]:
        result: List[Component] = []

        for suffix in [".jsonl", ".cmd"]:

            for file in path.rglob(f"*{suffix}"):
                if not file.is_file():
                    continue

                content = file.read_text()

                name_parts = []
                if len(prefix) > 0:
                    name_parts.append(prefix)
                name_parts = name_parts + list(file.relative_to(path).parts)

                name = "_".join(name_parts)
                component = Component(
                    name=name,
                    type=suffix[1:],
                    path=file,
                    content=content,
                )
                result.append(component)

        return result

    def _read_config(self, device_path: Path) -> Config | None:
        config_file = Path(device_path, CONFIG_FILE_NAME)
        if config_file.exists() and config_file.is_file():
            content = config_file.read_text()
            loaded = json.loads(content)

            config = Config(
                openhasp_config_manager=self._parse_openhasp_config_manager_config(loaded["openhasp_config_manager"]),
                mqtt=self._parse_mqtt_config(loaded["mqtt"]),
                http=self._parse_http_config(loaded["http"]),
                gui=self._parse_gui_config(loaded["gui"]),
                hasp=self._parse_hasp_config(loaded["hasp"])
            )

            return config

    @staticmethod
    def _parse_openhasp_config_manager_config(data: dict) -> OpenhaspConfigManagerConfig:
        return OpenhaspConfigManagerConfig(
            device=DeviceConfig(
                ip=data["device"]["ip"],
                screen=ScreenConfig(
                    width=data["device"]["screen"]["width"],
                    height=data["device"]["screen"]["height"]
                )
            )
        )

    @staticmethod
    def _parse_mqtt_config(data: dict) -> MqttConfig:
        return MqttConfig(
            name=data["name"],
            group=data["group"],
            host=data["host"],
            port=data["port"],
            user=data["user"],
            password=data["pass"],
        )

    @staticmethod
    def _parse_http_config(data: dict) -> HttpConfig:
        return HttpConfig(
            port=data["port"],
            user=data["user"],
            password=data["pass"],
        )

    @staticmethod
    def _parse_gui_config(data: dict) -> GuiConfig:
        return GuiConfig(
            idle1=data["idle1"],
            idle2=data["idle2"],
            bckl=data["bckl"],
            bcklinv=data["bcklinv"],
            rotate=data["rotate"],
            cursor=data["cursor"],
            invert=data["invert"],
            calibration=data["calibration"],
        )

    @staticmethod
    def _parse_hasp_config(data: dict) -> HaspConfig:
        return HaspConfig(
            startpage=data["startpage"],
            startdim=data["startdim"],
            theme=data["theme"],
            color1=data["color1"],
            color2=data["color2"],
            font=data["font"],
            pages=data["pages"],
        )

    def process(self, devices: List[Device] = None) -> List[Device]:
        """
        Process the configuration and generate the corresponding output
        :return:
        """
        if devices is None:
            devices = self.analyze()

        self._generate_output(devices)
        return devices

    def _generate_output(self, devices: List[Device]):
        for device in devices:

            self._clear_output(device)

            jsonl_processor = JsonlObjectProcessor()
            device_processor = DeviceProcessor(device, jsonl_processor, self._variable_manager)

            jsonl_validator = JsonlObjectValidator()
            device_validator = DeviceValidator(device.config, jsonl_validator)

            # "fill" the processor with all available data for this device
            for component in device.components:
                if component.type == "jsonl":
                    device_processor.add_jsonl(component)
                else:
                    device_processor.add_other(component)

            # let the processor manage each component
            for component in device.components:
                output_content = device_processor.normalize(component)
                try:
                    device_validator.validate(component, output_content)
                except Exception as ex:
                    print(f"Validation for {component.path} failed: {ex}")
                    raise ex

                self._write_output(device, component, output_content)

    @staticmethod
    def _write_output(device: Device, component: Component, output_content: str):
        device.output_dir.mkdir(parents=True, exist_ok=True)

        component_output_file = Path(
            device.output_dir,
            component.name
        )

        component_output_file.write_text(output_content)

    @staticmethod
    def _clear_output(device: Device):
        if device.output_dir.exists():
            shutil.rmtree(device.output_dir)



import json
import re
from typing import List, Dict

import jinja2
from jinja2 import Template

from openhasp_config_manager.model import Config, Component, Device
from openhasp_config_manager.processing.jsonl import JsonlObjectProcessor
from openhasp_config_manager.processing.variables import VariableManager


class DeviceProcessor:
    """
    A device specific processor, used to transform any templates
    present within the configuration files.
    """

    def __init__(self, device: Device, jsonl_object_processor: JsonlObjectProcessor, variable_manager: VariableManager):
        self._device = device

        self._id_object_map: Dict[str, dict] = {}
        self._others: List[Component] = []

        self._component_tree_changed = True
        self._template_vars: Dict[str, any] = {}

        self._jsonl_object_processor = jsonl_object_processor
        self._variable_manager = variable_manager

    def add_other(self, component: Component):
        self._others.append(component)
        self._component_tree_changed = True

    def add_jsonl(self, component: Component):
        parts = self._split_jsonl_objects(component.content)

        for part in parts:
            loaded = json.loads(part)

            object_key = f"p{loaded.get('page', '0')}b{loaded.get('id', '0')}"

            self._id_object_map[object_key] = loaded

        self._component_tree_changed = True

    def normalize(self, component: Component) -> str:
        if self._component_tree_changed:
            self._component_tree_changed = False
            self._template_vars = self._compute_template_variables()

        if component.type == "jsonl":
            return self._normalize_jsonl(self._device.config, component)
        if component.type == "cmd":
            return self._normalize_cmd(self._device.config, component)
        else:
            # no changes necessary
            return component.content

    def _normalize_jsonl(self, config: Config, component: Component) -> str:
        normalized_objects: List[str] = []

        objects = self._split_jsonl_objects(component.content)
        for ob in objects:
            p = self._normalize_jsonl_object(config, component, ob)
            normalized_objects.append(p)

        return "\n".join(normalized_objects)

    @staticmethod
    def _split_jsonl_objects(original_content: str) -> List[str]:
        pattern_to_find_beginning_of_objects = re.compile(r'^(?!\n)\s*(?=\{)', re.RegexFlag.MULTILINE)
        parts = pattern_to_find_beginning_of_objects.split(original_content)

        result = []
        for part in parts:
            part = part.strip()

            # edge case for first match
            if "}" not in part:
                continue

            # ignore lines starting with "//"
            part = "\n".join([line for line in part.splitlines() if not line.strip().startswith("//")])

            # ignore everything after the last closing bracket
            part = part.rsplit("}", maxsplit=1)[0] + "}"

            result.append(part)

        return result

    def _normalize_jsonl_object(self, config: Config, component: Component, ob: str) -> str:
        parsed = json.loads(ob)

        normalized_object = {}
        for key, value in parsed.items():
            if isinstance(value, str):
                template = Template(value)
                normalized_object[key] = template.render(self._template_vars)
            else:
                normalized_object[key] = value

        processed = self._jsonl_object_processor.process(normalized_object, config)
        return json.dumps(processed, indent=None)

    def _normalize_cmd(self, _device_config, component) -> str:
        template = Template(component.content)
        return template.render(self._template_vars)

    def _compute_template_variables(self) -> dict:
        result = {}

        # device specific variables
        result["device"] = self._device.config.openhasp_config_manager.device

        # object specific variables
        for key, obj in self._id_object_map.items():
            result[key] = obj

        result |= self._variable_manager.get_vars(None)
        result |= self._variable_manager.get_vars(self._device.name)

        rendered = self._render_dict_recursively(result)

        return rendered

    def _render_dict_recursively(self, input: Dict) -> Dict[str, any]:
        result = input
        last_successful_renders: int or None = None
        changed = True
        while changed:
            successful_renders = 0
            changed = False
            tmp = {}
            for key, value in result.items():

                rendered_key = key
                try:
                    rendered_key = Template(key).render(result)
                    changed = changed and rendered_key != key
                    successful_renders += 1
                except jinja2.UndefinedError as ex:
                    changed = True

                rendered_val = value
                if isinstance(value, str):
                    try:
                        rendered_val = Template(value).render(result)
                        successful_renders += 1
                    except jinja2.UndefinedError as ex:
                        changed = True
                if isinstance(value, dict):
                    try:
                        rendered_val = self._render_dict_recursively(value)
                        successful_renders += 1
                    except jinja2.UndefinedError as ex:
                        changed = True

                changed = changed and rendered_val != value
                tmp[rendered_key] = rendered_val

            result = tmp

            if last_successful_renders is not None and last_successful_renders == successful_renders:
                raise jinja2.UndefinedError("Unable to progress while rendering")
            last_successful_renders = successful_renders

        return result



import json
from typing import Dict, List

import requests as requests

from openhasp_config_manager.model import Device, MqttConfig, HttpConfig, GuiConfig, HaspConfig

GET = "GET"
POST = "POST"
DELETE = "DELETE"


class OpenHaspClient:

    def get_files(self, device: Device) -> List[str]:
        """
        Retrieve a list of all file on the device
        :param device: the device to query
        :return: a list of all files on the device
        """
        base_url = self._compute_base_url(device)

        username = device.config.http.user
        password = device.config.http.password

        response = self._do_request(
            GET,
            base_url + "list?dir=/",
            username=username, password=password
        )
        response_data = json.loads(response.decode('utf-8'))

        files = list(filter(lambda x: x["type"] == "file", response_data))
        file_names = list(map(lambda x: x["name"], files))
        return file_names

    def get_file_content(self, device: Device, file_name: str) -> str or None:
        base_url = self._compute_base_url(device)

        username = device.config.http.user
        password = device.config.http.password

        try:
            response = self._do_request(
                GET,
                base_url + file_name,
                username=username, password=password
            )
            response_data = response.decode('utf-8')
            return response_data
        except Exception as ex:
            return None

    def delete_file(self, device: Device, file_name: str):
        """
        Delete a file on the device
        :param device: the target device
        :param file_name: the name of the file
        """
        base_url = self._compute_base_url(device)

        username = device.config.http.user
        password = device.config.http.password

        self._do_request(
            DELETE,
            base_url + "edit",
            data={
                "path": "/" + file_name
            },
            username=username, password=password
        )

    def command(self, device: Device, name: str, params: str):
        """
        Execute a command on a device
        :param device: the device to request
        :param name: the name of the command
        :param params: parameters for the command
        """

        import paho.mqtt.client as paho

        mqtt_host = device.config.mqtt.host
        mqtt_port = device.config.mqtt.port
        mqtt_client_id = 'openhasp-config-manager'
        mqtt_user = device.config.mqtt.user
        mqtt_password = device.config.mqtt.password

        topic = f"hasp/{device.config.mqtt.name}/command/{name}"

        client = paho.Client(client_id=mqtt_client_id, protocol=paho.MQTTv5)

        client.username_pw_set(username=mqtt_user, password=mqtt_password)
        client.connect(mqtt_host, mqtt_port)

        data = params.strip('"')
        result = client.publish(topic=topic, payload=data)
        result.wait_for_publish()

        if not result.rc == paho.MQTT_ERR_SUCCESS:
            print('Code %d while sending message %d: %s' % (result.rc, result.mid, paho.error_string(result.rc)))

    def reboot(self, device: Device):
        """
        Request a reboot
        :param device: the target device
        """
        base_url = self._compute_base_url(device)
        username = device.config.http.user
        password = device.config.http.password
        self._do_request(
            GET, base_url + "reboot",
            username=username, password=password
        )

    def set_hasp_config(self, device: Device, config: HaspConfig):
        """
        Set the "HASP" configuration
        :param device: the target device
        :param config: the configuration to set
        """
        base_url = self._compute_base_url(device)

        data = {
            "startpage": config.startpage,
            "startdim": config.startdim,
            "theme": config.theme,
            "color1": config.color1,
            "color2": config.color2,
            "font": config.font,
            "pages": config.pages,
            "save": "hasp"
        }

        # ignore keys with None value
        data = {k: v for k, v in data.items() if v is not None}

        username = device.config.http.user
        password = device.config.http.password

        self._do_request(
            POST, base_url + "config",
            data=data,
            username=username, password=password
        )

    def set_http_config(self, device: Device, config: HttpConfig):
        """
        Set the HTTP configuration
        :param device: the target device
        :param config: the configuration to set
        """
        base_url = self._compute_base_url(device)

        data = {
            "user": config.user,
            "pass": config.password,
            "save": "http"
        }

        # ignore keys with None value
        data = {k: v for k, v in data.items() if v is not None}

        username = device.config.http.user
        password = device.config.http.password

        self._do_request(
            POST, base_url + "config",
            data=data,
            username=username, password=password
        )

    def set_mqtt_config(self, device: Device, config: MqttConfig):
        """
        Set the MQTT configuration
        :param device: the target device
        :param config: the configuration to set
        :return:
        """
        base_url = self._compute_base_url(device)

        data = {
            "name": config.name,
            "group": config.group,
            "host": config.host,
            "port": config.port,
            "user": config.user,
            "pass": config.password,
            "save": "mqtt"
        }

        # ignore keys with None value
        data = {k: v for k, v in data.items() if v is not None}

        username = device.config.http.user
        password = device.config.http.password

        self._do_request(
            POST, base_url + "config",
            data=data,
            username=username, password=password
        )

    def set_gui_config(self, device: Device, config: GuiConfig):
        """
        Set the GUI configuration
        :param device: the target device
        :param config: the configuration to set
        :return:
        """
        base_url = self._compute_base_url(device)

        data = {
            "idle1": config.idle1,
            "idle2": config.idle2,
            "rotate": config.rotate,
            "cursor": config.cursor,
            "bckl": config.bckl,
            "save": "gui"
        }

        # ignore keys with None value
        data = {k: v for k, v in data.items() if v is not None}

        username = device.config.http.user
        password = device.config.http.password

        self._do_request(
            POST, base_url + "config",
            data=data,
            username=username, password=password
        )

    def upload_files(self, device: Device, files: Dict[str, str]):
        """
        Upload a collection of files
        :param device: the target device
        :param files: "target file name"->"file content" mapping
        """
        for name, content in files.items():
            self.upload_file(device, name, content)

    def upload_file(self, device: Device, name: str, content: str):
        """
        Upload a single file
        :param device: the target device
        :param name: the target name of the file on the device
        :param content: the file content
        """
        print(f"Uploading '{name}'...")

        url = self._compute_base_url(device)
        url += "edit"

        username = device.config.http.user
        password = device.config.http.password
        self._do_request(
            POST, url,
            files={
                f"{name}": content
            },
            username=username, password=password
        )

    @staticmethod
    def _do_request(method: str = GET, url: str = "/", params: dict = None,
                    json: any = None, files: any = None, data: any = None,
                    headers: dict = None,
                    username: str = None, password: str = None) -> list or dict or None:
        """
        Executes a http request based on the given parameters

        :param method: the method to use (GET, POST)
        :param url: the url to use
        :param params: query parameters that will be appended to the url
        :param json: request body
        :param headers: custom headers
        :return: the response parsed as a json
        """
        _headers = {
        }
        if headers is not None:
            _headers.update(headers)

        response = requests.request(
            method, url, headers=_headers,
            json=json, files=files,
            data=data,
            auth=(username, password),
            timeout=5
        )

        response.raise_for_status()
        if len(response.content) > 0:
            if "application/json" in response.headers.get("Content-Type", ""):
                return response.json()
            else:
                return response.content

    @staticmethod
    def _compute_base_url(device: Device) -> str:
        url = device.config.openhasp_config_manager.device.ip
        if not url.startswith("http://"):
            url = "http://" + url
        if not url.endswith("/"):
            url += "/"
        return url



import re
from typing import Dict

from openhasp_config_manager.model import Config


class JsonlObjectProcessor:
    PERCENTAGE_REGEX_PATTERN = re.compile(r"^\d+(\.\d+)?%$")

    def process(self, input: Dict, config: Config) -> Dict:
        result: Dict[str, any] = {}
        for key, value in input.items():
            if isinstance(value, str) and re.match(self.PERCENTAGE_REGEX_PATTERN, value):
                numeric_value = self._parse_percentage(value)

                total_width = config.openhasp_config_manager.device.screen.width
                total_height = config.openhasp_config_manager.device.screen.height

                # switch width and height values if screen is rotated an odd amount of times
                if config.gui.rotate % 2 == 1:
                    t = total_width
                    total_width = total_height
                    total_height = t

                if key in ["x", "w"]:
                    total = total_width
                else:
                    total = total_height

                result[key] = self._percentage_of(numeric_value, total)
            else:
                result[key] = value

        # normalize value types, in case of templates
        for key, value in result.items():
            if key in [
                "page", "id",
                "x", "y", "w", "h",
                "text_font", "value_font",
                "radius", "border_side",
                "min", "max",
                "prev", "next",
            ] and isinstance(value, str):
                result[key] = int(float(value))

        return result

    @staticmethod
    def _percentage_of(percentage: float, total: int) -> int:
        """
        :param percentage: 0..100
        :param total: value for a percentage of 100
        :return: value according to input
        """
        return int((percentage * total) / 100)

    @staticmethod
    def _parse_percentage(value: str) -> float:
        return float(str(value).replace('%', ''))



from pathlib import Path
from typing import Dict

import yaml
from yaml import Loader

from openhasp_config_manager.const import COMMON_FOLDER_NAME, DEVICES_FOLDER_NAME


class VariableManager:
    global_vars = {}
    device_vars = {}

    def __init__(self, cfg_root: Path):
        self.cfg_root = Path(cfg_root)
        self._read()

    def get_vars(self, device: str | None) -> Dict:
        if device is not None:
            return self.device_vars.get(device, {})
        else:
            return self.global_vars

    def _read(self):

        global_var_files = self.cfg_root.rglob(f"{COMMON_FOLDER_NAME}/*.yaml")
        for file in global_var_files:
            data = self._load_var_file(file)
            self.global_vars |= data

        devices_path = Path(self.cfg_root, DEVICES_FOLDER_NAME)
        for device_dir in devices_path.iterdir():
            if not device_dir.is_dir():
                continue

            device_var_files = device_dir.rglob(f"*.yaml")
            for file in device_var_files:
                if not file.is_file():
                    continue

                data = self._load_var_file(file)

                device_name = file.parent.name
                if device_name not in self.device_vars.keys():
                    self.device_vars[device_name] = {}

                self.device_vars[device_name] |= data

    @staticmethod
    def _load_var_file(file: Path):
        content = file.read_text()
        return yaml.load(content, Loader=Loader)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['openhasp_config_manager',
 'openhasp_config_manager.model',
 'openhasp_config_manager.processing',
 'openhasp_config_manager.validation']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2', 'click', 'paho-mqtt', 'py-range-parse', 'pyyaml', 'requests']

entry_points = \
{'console_scripts': ['openhasp-config-manager = '
                     'openhasp_config_manager.cli:cli']}

setup_kwargs = {
    'name': 'openhasp-config-manager',
    'version': '0.0.3',
    'description': 'A tool to manage all of your openHASP device configs in a centralized place.',
    'long_description': '# openhasp-config-manager\n\nA tool to manage all of your [openHASP](https://github.com/HASwitchPlate/openHASP) device configs in a centralized place.\n\n# Features\n\n* [x] manage as many devices as you like\n* [x] share configuration files between devices\n* [x] jsonl preprocessing, which allows for\n  * [x] `//` comments within jsonl files\n  * [x] line breaks wherever you like\n  * [x] jinja2 templating within object values\n* [x] output validation\n  * [x] no more weird behavior due to invalid "id" range\n* [x] one click configuration upload to the device\n  * [x] automatic diffing to only update changed configuration files\n* [x] execute commands directly from the CLI (still needs a connection to the MQTT broker)\n\n# How to use\n\n## Installation\n\nSince openhasp-config-manager needs some dependencies (see [here](/pyproject.toml)) it is\nrecommended to install it inside of a virtualenv:\n\n```bash\nmkdir -p ~/venvs/openhasp-config-manager\npython3 -m venv ~/venvs/openhasp-config-manager\nsource ~/venvs/openhasp-config-manager/bin/activate\npip3 install openhasp-config-manager\nopenhasp-config-manager -h\n```\n\n## Configuration\n\nopenhasp-config-manager is first and foremost a configuration\nmanagement system. Simply follow the basic folder structure and\nconfig deployment will become trivial.\n\n* `devices`: In the root directory of your configuration, a folder called\n  `devices` is expected.\n  * In there you can create as many subfolders as\n    you like, naming them according to the physical devices that you\n    want to manage.\n    * Within those device subfolders you can then create\n      `*.jsonl` and `*.cmd` files.\n    * You must also provide a `config.json` file, see [config.json](#config.json)\n      for more info on how to set it.\n* `common`: The `common` directory can be used to put files\n  that should be included on _all_ device.\n\nYou are not limited to a folder depth of one. However, the files\non OpenHasp devices cannot be put into subfolders. Therefore, if you put\n`.json` or `.cmd` files into subfolders, the name of the\nresulting file on the OpenHasp device will be a concatenation of\nthe full subpath using an underscore (`_`) as a separator. So f.ex.\nthe file in the following structure:\n\n```text\nopenhasp-configs\n└── devices\n    └── touch_down_1\n        └── 0_home\n            └── 0_header.jsonl\n```\n\nwould only be uploaded to the `touch_down_1` device and named:\n`0_home_0_header.jsonl`\n\nA more advanced configuration layout could look something like this:\n\n```text\nopenhasp-configs\n├── common\n│\xa0\xa0 ├── content\n│\xa0\xa0 │\xa0\xa0 └── card.jsonl\n│\xa0\xa0 ├── dialog\n│\xa0\xa0 │\xa0\xa0 ├── connected.jsonl\n│\xa0\xa0 │\xa0\xa0 └── offline.jsonl\n│\xa0\xa0 ├── navigation_bar.jsonl\n│\xa0\xa0 └── page_header.jsonl\n└── devices\n    └── touch_down_1\n        ├── 0_home\n        │\xa0\xa0 ├── 0_header.jsonl\n        │\xa0\xa0 ├── 1_content.jsonl\n        │\xa0\xa0 └── page.cmd\n        ├── 5_about\n        │\xa0\xa0 ├── 0_header.jsonl\n        │\xa0\xa0 ├── 1_content.jsonl\n        │\xa0\xa0 └── page.cmd\n        ├── boot.cmd\n        ├── config.json\n        ├── offline.cmd\n        └── online.cmd\n```\n\n### config.json\n\nopenhasp-config-manager makes use of the `config.json` on your plate. It can use information \nto detect things like screen orientation, and also allows you to deploy config changes to your\nplate when you make changes in the config.json file. Since [the official API does not support\nuploading the full file](https://github.com/HASwitchPlate/openHASP/issues/363), only settings \nwhich can also be set through the web ui on the plate itself are currently supported.\n\nTo retrieve the initial version of the `config.json` file you can use the\nbuilt-in file browser integrated into the webserver of your OpenHASP plate.\n\nThe official `config.json` file doesn\'t provide enough info for openhasp-config-manager\nto enable all of its features though. To fix that openhasp-config-manager looks for an \nadditional section within the file which is not present by default:\n\n```json\n{\n  "openhasp_config_manager": {\n    "device": {\n      "ip": "192.168.5.134",\n      "screen": {\n        "width": 320,\n        "height": 480\n      }\n    }\n  },\n  "wifi": {\n    "ssid": "Turris IoT",\n    ...\n  }\n```\n\nSimply add this section to the `config.json` after you have retrieved it from\nthe plate.\n\n### Preprocessing\n\nopenhasp-config-manager runs all configuration files through a preprocessor, which allows us to use\nfeatures the original file format doesn\'t support, like f.ex. templating.\n\n#### Templating\n\nYou can use Jinja2 templates inside of values. You can access each of the objects using the\n`pXbY` syntax established by OpenHASP, where `X` is the `page` of an object and `Y` is its `id`.\n\nYou can use the full functionality of Jinja2 like f.ex. math operations, function calls or type conversions.\n\n```yaml\n{\n  "page": 1,\n  "id": 1,\n  "x": 0,\n  "y": 0,\n  ...\n}\n\n{\n  "page": 1,\n  "id": 2,\n  "x": "{{ p1b1.x }}",\n  "y": "{{ p1b1.y + 10 }}",\n  ...\n}\n```\n\n#### Variables\n\nBesides accessing other objects, you can also define custom variables yourself, which can then\nbe used inside of templates.\n\n##### Global\n\nGlobal variables can be specified by creating `*.yaml` files inside of the `common` folder.\n\nExample:\n\n`common/global.vars.yaml`\n\n```yaml\nabout:\n  page_title: "About"\n```\n\n`common/about_page.jsonl`\n\n```json lines\n{\n  "page": 9,\n  "id": 1,\n  ...\n  "title": "{{ about.page_title }}",\n  ...\n}\n```\n\n##### Device specific\n\nDevice specific variables can be specified by creating `*.yaml` files inside any of the sub-folders\nof the `device` folder.\n\n> **Note**\n>\n> Device specific variables will override global variables, given the same name.\n\nExample:\n\n`device/my_device/device.vars.yaml`\n\n```yaml\npage_title: "My Device"\n```\n\n`device/my_device/some_folder/some_page.jsonl`\n\n```json lines\n{\n  "page": 1,\n  "id": 1,\n  ...\n  "title": "{{ page_title }}",\n  ...\n}\n```\n\n`device/my_device/some_other_folder/some_page.jsonl`\n\n```json lines\n{\n  "page": 2,\n  "id": 1,\n  ...\n  "title": "{{ page_title }}",\n  ...\n}\n```\n\n## Deployment\n\nTo deploy your configurations to the already connected OpenHASP devices, simply use the\ncommand line tool `openhasp-config-manager`:\n\n```shell\n> openhasp-config-manager                                                         0 (0.604s) < 02:11:38\nUsage: openhasp-config-manager [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --version   Show the version and exit.\n  -h, --help  Show this message and exit.\n\nCommands:\n  cmd       Sends a command request to a device.\n  deploy    Combines the generation and upload of a configuration.\n  generate  Generates the output files for all devices in the given...\n  upload    Uploads the previously generated configuration to their...\n```\n\n## Run commands\n\nOpenHasp allows you to run commands on a device by issuing MQTT messages.\nWhile openhasp-config-manager is first and foremost a config management system,\nit also allows you to run commands on a device without the need to install a separate\nMQTT client first. Note that the MQTT _server_ does need to be running and also has to\nbe reachable from your local machine.\n\nSee: https://openhasp.haswitchplate.com/latest/commands/\n\n```shell\n> openhasp-config-manager cmd -c ./openhasp-configs -d plate35 -C backlight -p "{\\"state\\":\\"on\\",\\"brightness\\":128}"\n```\n\n# FAQ\n\n## How do I see device logs?\n\nWith the device connected via USB cable, open a terminal and run:\n\n```shell\nbash -c "screen -q -L -Logfile device.log /dev/ttyUSB0 115200 &> /dev/null; tail -F device.log; killall screen"\n```\n\n# Contributing\n\nGitHub is for social coding: if you want to write code, I encourage contributions\nthrough pull requests from forks of this repository. Create GitHub tickets for\nbugs and new features and comment on the ones that you are interested in.\n\n# License\n\n```text\nopenhasp-config-manager is free software: you can redistribute it and/or modify\nit under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE as published by\nthe Free Software Foundation, either version 3 of the License, or\n(at your option) any later version.\n\nThis program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\nGNU General Public License for more details.\n\nYou should have received a copy of the GNU General Public License\nalong with this program.  If not, see <https://www.gnu.org/licenses/>.\n```\n',
    'author': 'Markus Ressel',
    'author_email': 'mail@markusressel.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/markusressel/openhasp-config-manager',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
